/*
 * Copyright (C) 2018 CERN (www.cern.ch)
 * Author: Jean-Claude BAU & Maciej Lipinski
 *
 * Released according to the GNU LGPL, version 2.1 or any later version.
 */

#include <ppsi/ppsi.h>
#include <common-fun.h>
#include "l1e-api.h"
#include "l1e-constants.h"
#include <math.h>

typedef struct {
	int (*action)(struct pp_instance *ppi, Boolean newState);
}l1e_state_machine_t;

#define L1E_DEFAULT_DELAY 500

extern char *l1e_state_name[];

#define MAX_STATE_ACTIONS (sizeof(le1_state_actions)/sizeof(l1e_state_machine_t))

/* prototypes */
static int l1e_empty_action(struct pp_instance *ppi, Boolean new_state);
static int l1e_handle_state_disabled(struct pp_instance *ppi, Boolean new_state);
static int l1e_handle_state_idle(struct pp_instance *ppi, Boolean new_state);
static int l1e_handle_state_link_alive(struct pp_instance *ppi, Boolean new_state);
static int l1e_handle_state_config_match(struct pp_instance *ppi, Boolean new_state);
static int l1e_handle_state_up(struct pp_instance *ppi, Boolean new_state);

static l1e_state_machine_t le1_state_actions[] ={
		[0] { /* Not used */
			.action=l1e_empty_action,
		},
		[L1SYNC_DISABLED] {
			.action=l1e_handle_state_disabled,
		},
		[L1SYNC_IDLE] {
			.action=l1e_handle_state_idle,
		},
		[L1SYNC_LINK_ALIVE] {
			.action=l1e_handle_state_link_alive,
		},
		[L1SYNC_CONFIG_MATCH] {
			.action=l1e_handle_state_config_match,
		},
		[L1SYNC_UP] {
			.action=l1e_handle_state_up,
		},
};


/*
 * This hook is called by fsm to run the extension state machine.
 * It is used to send signaling messages.
 * It returns the ext-specific timeout value
 */
int l1e_run_state_machine(struct pp_instance *ppi) {
	L1SyncBasicPortDS_t * basicDS=L1E_DSPOR_BS(ppi);

	Enumeration8 nextState=basicDS->next_state;
	Boolean newState=nextState!=basicDS->L1SyncState;
	int delay;

	if ( nextState>=MAX_STATE_ACTIONS)
		return L1E_DEFAULT_DELAY; /* return default delay */

	/*
	 *  Update the L1SYNC dynamic data independent of the state machine
	 */

	/* By design, when link is up, it is always tx coherent and congruent*/
	basicDS->isTxCoherent= ppi->link_up ? 1 : 0;

	/* Check SYNC reception Time-out */
	if ( pp_timeout(ppi, L1E_TIMEOUT_RX_SYNC) ) {
		/* Time-out detected */
		basicDS->L1SyncLinkAlive = FALSE;
	}

	/*
	 * Evaluation of events common to all states
	 */
	if ( newState ) {
		basicDS->L1SyncState=nextState;
		pp_diag(ppi, ext, 2, "L1SYNC state: Enter %s\n", l1e_state_name[nextState]);
	}
	delay=(*le1_state_actions[basicDS->L1SyncState].action) (ppi,newState);
	if ( basicDS->L1SyncState != basicDS->next_state )
		pp_diag(ppi, ext, 2, "L1SYNC state: Exit %s\n", l1e_state_name[basicDS->L1SyncState]);
	return delay;
}

static int l1e_empty_action(struct pp_instance *ppi, Boolean new_state){
	return L1E_DEFAULT_DELAY; /* return default delay */
}

/* L1_SYNC_RESET event */
static inline Boolean le1_evt_L1_SYNC_RESET(struct pp_instance *ppi) {
	return ppi->link_up == 0;
}

/* L1_SYNC_ENABLED event */
static inline Boolean le1_evt_L1_SYNC_ENABLED(struct pp_instance *ppi) {
	return L1E_DSPOR_BS(ppi)->L1SyncEnabled == TRUE;
}

/* CONFIG_OK event */
static inline Boolean le1_evt_CONFIG_OK(struct pp_instance *ppi) {
	L1SyncBasicPortDS_t *basicDS=L1E_DSPOR_BS(ppi);

	return basicDS->txCoherentIsRequired==basicDS->peerTxCoherentIsRequired
			&& basicDS->rxCoherentIsRequired==basicDS->peerRxCoherentIsRequired
			&& basicDS->congruentIsRequired==basicDS->peerCongruentIsRequired;
}


/* STATE_OK event */
static Boolean le1_evt_STATE_OK(struct pp_instance *ppi) {
	L1SyncBasicPortDS_t *basicDS=L1E_DSPOR_BS(ppi);
	int pll_state;

	/* Update isRxCoherent and isCongruant : it must follow the PTP state machine */
	switch (ppi->state) {
	case PPS_SLAVE :
	case PPS_UNCALIBRATED :
		pll_state= WRH_OPER()->locking_poll(ppi, 0); /* Get the PPL state */
		basicDS->isCongruent =
				basicDS->isRxCoherent= pll_state == WRH_SPLL_READY ? 1 : 0;
		break;
	case PPS_MASTER :
		basicDS->isRxCoherent=
					basicDS->peerIsTxCoherent  &&
					basicDS->peerIsRxCoherent  &&
					basicDS->isTxCoherent    ? 1 : 0;
		/** TODO: in principle, the PLL should report unlocked when it is not locked,
		 * regardless of the state. Thus, when we check in MASTER state, it should report
		 * unlocked - this is not the case for some reasons
		 * Ideally, we want to check whether PLL is locked at any time. This is to cover
		 * the case of network re-arrangement. In such case, a port in SLAVE state can
		 * transition to MASTER state and after such transition the port in MASTER state
		 * might remain locked. We want to detect this case.
		 *
		 * Thus, the condition should be:
		 * basicDS->isCongruent= pll_state == !WRH_SPLL_READY ? 1 : 0;
		 */
		basicDS->isCongruent = 1;
		break;
	default :
		basicDS->isCongruent = 0;
	}

	return ( !basicDS->txCoherentIsRequired  || (basicDS->isTxCoherent && basicDS->peerIsTxCoherent ))
		&& ( !basicDS->rxCoherentIsRequired  || (basicDS->isRxCoherent && basicDS->peerIsRxCoherent ))
		&& ( !basicDS->congruentIsRequired   || (basicDS->isCongruent && basicDS->peerIsCongruent));
}

/* LINK_OK event */
static inline Boolean le1_evt_LINK_OK(struct pp_instance *ppi) {
	return L1E_DSPOR(ppi)->basic.L1SyncLinkAlive == TRUE;
}

static void l1e_send_sync_msg(struct pp_instance *ppi, Boolean immediatSend) {
	if (pp_timeout(ppi, L1E_TIMEOUT_TX_SYNC) || immediatSend) {
		int len;

		pp_diag(ppi, ext, 1, "Sending L1SYNC_TLV signaling msg\n");
		len = l1e_pack_signal(ppi);
		/* FIXME: check the destination MAC address */
		__send_and_log(ppi, len, PP_NP_GEN);

		/* Calculate when the next message should be sent */
		__pp_timeout_set(ppi, L1E_TIMEOUT_TX_SYNC,
				4 << (L1E_DSPOR_BS(ppi)->logL1SyncInterval + 8)); /* loop ever since */
	}
}

/* DISABLED state */
static int l1e_handle_state_disabled(struct pp_instance *ppi, Boolean new_state){
	l1e_ext_portDS_t * l1e_portDS=L1E_DSPOR(ppi);
	/* State initialization */
	if ( new_state ) {
		/* Table 157 - page 449
		 * All dynamic members of L1SyncBasicPortDS and L1SyncOptPortDS data sets are set
		 * to initialization values
		 */
		l1e_portDS->basic.isTxCoherent =
				l1e_portDS->basic.isRxCoherent =
						l1e_portDS->basic.isCongruent =
								l1e_portDS->basic.peerTxCoherentIsRequired =
										l1e_portDS->basic.peerRxCoherentIsRequired =
												l1e_portDS->basic.peerCongruentIsRequired =
														l1e_portDS->basic.peerIsTxCoherent =
																l1e_portDS->basic.peerIsRxCoherent =
																		l1e_portDS->basic.peerIsCongruent =
																				FALSE;
		/* Set extension mode disabled */
		l1e_portDS->head.extModeOn = l1e_portDS->parentExtModeOn = 0;
	}
	/* Check if state transition needed */
	if ( le1_evt_L1_SYNC_ENABLED(ppi) && !le1_evt_L1_SYNC_RESET(ppi) ) {
		/* Go the IDLE state */
		 L1E_DSPOR_BS(ppi)->next_state=L1SYNC_IDLE;
		 return 0; /* no wait to evaluate next state */
	}
	return L1E_DEFAULT_DELAY; /* return default delay */
}

/* IDLE state */
static int l1e_handle_state_idle(struct pp_instance *ppi, Boolean new_state){
	l1e_ext_portDS_t * l1e_portDS=L1E_DSPOR(ppi);

	/* State initialization */
	if ( new_state ) {
		/* Table 157 - page 449
		 * The dynamic members listed in Table 155 are set to initialization
		 * values when entering this state.
		 */
		l1e_portDS->basic.peerTxCoherentIsRequired =
				l1e_portDS->basic.peerRxCoherentIsRequired =
						l1e_portDS->basic.peerCongruentIsRequired =
								l1e_portDS->basic.peerIsTxCoherent  =
										l1e_portDS->basic.peerIsRxCoherent=
												l1e_portDS->basic.peerIsCongruent = FALSE;
		/* Set extension mode disabled */
		l1e_portDS->head.extModeOn = l1e_portDS->parentExtModeOn = 0;
		l1e_send_sync_msg(ppi,1); /* Send immediately a message */
	}

	/* Check if state transition needed */
	if ( !le1_evt_L1_SYNC_ENABLED(ppi) || le1_evt_L1_SYNC_RESET(ppi) ) {
		/* Go to DISABLE state */
		l1e_portDS->basic.next_state=L1SYNC_DISABLED;
		return L1E_DEFAULT_DELAY; /* return default delay */
	}
	if ( le1_evt_LINK_OK(ppi) ) {
		l1e_portDS->basic.next_state=L1SYNC_LINK_ALIVE;
 		return L1E_DEFAULT_DELAY; /* return default delay */
	}
	/* Iterative treatment */
	l1e_send_sync_msg(ppi,0);
	return L1E_DEFAULT_DELAY; /* return default delay */
}

/* LINK_ALIVE state */
static int l1e_handle_state_link_alive(struct pp_instance *ppi, Boolean new_state){
	L1SyncBasicPortDS_t *basic= L1E_DSPOR_BS(ppi);

	/* State initialization */
	if ( new_state ) {
		/* Initialize time-out peer L1SYNC reception */
		int timeout_tx_sync = (4 << (basic->logL1SyncInterval + 8)) * basic->L1SyncReceiptTimeout;
		__pp_timeout_set(ppi, L1E_TIMEOUT_RX_SYNC, timeout_tx_sync);

	}
	/* Check if state transition needed */
	if ( !le1_evt_LINK_OK(ppi) ) {
		/* Go to IDLE state */
		basic->next_state=L1SYNC_IDLE;
		return L1E_DEFAULT_DELAY; /* return default delay */
	}
	/* we will allow to go to CONFIG_MATH only for UNCALIBRATED and MASTER states */
	if ( le1_evt_CONFIG_OK(ppi) && (ppi->state==PPS_UNCALIBRATED || ppi->state==PPS_MASTER)) {
		basic->next_state=L1SYNC_CONFIG_MATCH;
		return 0; /* Try to treat next state asap */
	}
	/* Iterative treatment */
	l1e_send_sync_msg(ppi,0);
	return  pp_next_delay_1(ppi, L1E_TIMEOUT_TX_SYNC); /* Return the timeout for next invocation */
}

/* CONFIG_MATCH state */
static int l1e_handle_state_config_match(struct pp_instance *ppi, Boolean new_state){
	L1SyncBasicPortDS_t *basic= L1E_DSPOR_BS(ppi);

	/* State initialization */
	if (new_state) {
		switch ( ppi->state ) {
			case PPS_SLAVE :
			case PPS_UNCALIBRATED :
				if ( basic->congruentIsRequired==1 &&	basic->isRxCoherent==0) {
					pp_diag(ppi, ext, 1, "Locking PLL\n");
					WRH_OPER()->locking_enable(ppi);
				}
				break;
			case PPS_MASTER :
				if (  basic->congruentIsRequired == 1) {
					WRH_OPER()->locking_disable(ppi);
				}
				break;
			default:
				break;
		}
	}
	/* Check if state transition needed */
	if ( !le1_evt_LINK_OK(ppi) ) {
		/* Go to IDLE state */
		L1E_DSPOR_BS(ppi)->next_state=L1SYNC_IDLE;
		return L1E_DEFAULT_DELAY; /* return default delay */
	}
	if ( !le1_evt_CONFIG_OK(ppi) ) {
		/* Return to LINK_ALIVE state */
		L1E_DSPOR_BS(ppi)->next_state=L1SYNC_LINK_ALIVE;
		return L1E_DEFAULT_DELAY; /* return default delay */
	}
	if ( le1_evt_STATE_OK(ppi) ) {
		/* Return to UP state */
		L1E_DSPOR_BS(ppi)->next_state=L1SYNC_UP;
		return 0; /* Treat next state asap */
	}
	/* Iterative treatment */
	l1e_send_sync_msg(ppi,0);
	return  pp_next_delay_1(ppi, L1E_TIMEOUT_TX_SYNC); /* Return the timeout for next invocation */
}

/* UP state */
static int l1e_handle_state_up(struct pp_instance *ppi, Boolean new_state){
	l1e_ext_portDS_t * l1e_portDS=L1E_DSPOR(ppi);
	Enumeration8 next_state=0;

	/* State initialization */
	if ( new_state ) {
		l1e_servo_init(ppi); /* The servo can be initialized because the PPL is locked */
		l1e_portDS->head.extModeOn =
				l1e_portDS->parentExtModeOn = 1;
		WRH_OPER()->enable_ptracker(ppi);
	}

	/* Check if state transition needed */
	if ( !le1_evt_LINK_OK(ppi) ) {
		/* Go to IDLE state */
		next_state=L1SYNC_IDLE;
	}
	if ( !le1_evt_CONFIG_OK(ppi) ) {
		/* Return to LINK_ALIVE state */
		next_state=L1SYNC_LINK_ALIVE;
	}
	if ( !le1_evt_STATE_OK(ppi) ) {
		/* Return to CONFIG_MATCH state */
		next_state=L1SYNC_CONFIG_MATCH;
	}
	if (next_state!=0 ) {
		l1e_portDS->basic.next_state=next_state;
		l1e_portDS->head.extModeOn =
				l1e_portDS->parentExtModeOn = 0;
		WRH_OPER()->locking_disable(ppi); /* Unlock the PLL */
		l1e_servo_reset(ppi);
		return 0; /* Treat the next state asap */
	}

	/* Iterative treatment */
	l1e_update_correction_values(ppi);
	l1e_send_sync_msg(ppi,0);
	return  pp_next_delay_1(ppi, L1E_TIMEOUT_TX_SYNC); /* Return the timeout for next invocation */
}
